#!/usr/bin/env python3

import appcli
import autoprop
from inform import warn
from appcli import Key, DocoptConfig
from dataclasses import dataclass
from stepwise_mol_bio import Main, UsageError

FRAGMENT_DOC = """\
    <backbone> <inserts>
        The DNA fragments to assemble.  For each fragment, the following 
        information can be specified:
        
        - A name (optional): This is how the fragment will be referred to in 
          the protocol, but will not have any effect other than that.  By 
          default, a generic name like "Insert #1" will be chosen.

        - A concentration (required): This is used to calculate how much of 
          each fragment needs to be added to get the ideal backbone:insert 
          ratio.  You may specify a unit (e.g. nM), but ng/µL will be assumed 
          if you don't.  You may also specify the concentration as just "PCR" 
          (no other value or units).  This indicates that the fragment is the 
          product of an PCR reaction, and is assumed to be 50 ng/μL.

        - A length (required if concentration is in ng/µL): The length of the 
          fragment in bp.  This is used to convert the above concentration into
          a molarity.

        For each individual fragment, specify whichever of the above fields are 
        relevant, separated by colons.  In other words:
        
            [<name>:]<conc>[:<length>]"""

OPTION_DOC = """\
    -n --num-reactions <int>        [default: ${app.num_reactions}]
        The number of reactions to setup.

    -m, --master-mix <bb,ins>       [default: ${app.master_mix_str}]
        Indicate which fragments should be included in the master mix.  Valid 
        fragments are "bb" (for the backbone), "ins" (for all the inserts), 
        "1" (for the first insert), "2", "3", etc.

    -v, --reaction-volume <µL>      [default: ${app.volume_uL}]
        The volume of the complete assembly reaction.  You might want larger 
        reaction volumes if your DNA is dilute, or if you have a large number 
        of inserts.

    -x, --excess-insert <ratio>     [default: ${app.excess_insert}]
        The molar-excess of each insert relative to the backbone.  Values 
        between 1-10 (e.g. 1-10x excess) are typical."""

class Assembly(Main):
    __config__ = [
            DocoptConfig(
                usage_getter=lambda self: self._get_docopt_usage(),
            ),
    ]

    fragments = appcli.param(
            lambda d: parse_fragments([d['<backbone>'], *d['<inserts>']]),
    )
    num_reactions = appcli.param(
            Key(DocoptConfig, '--num-reactions'),
            cast=int,
            default=1,
    )
    volume_uL = appcli.param(
            Key(DocoptConfig, '--reaction-volume'),
            cast=float,
            default=5,
    )
    master_mix = appcli.param(
            Key(DocoptConfig, '--master-mix'),
            cast=lambda x: frozenset(x.split(',')),
            default=frozenset(),
    )
    excess_insert = appcli.param(
            Key(DocoptConfig, '--excess-insert'),
            cast=float,
            default=2,
    )

    target_pmol_per_frag = 0.06
    min_pmol_per_frag = 0.02

    @property
    def master_mix_str(self):
        return ','.join(self.master_mix)

    def _get_docopt_usage(self):
        return self.__doc__.format(self, **globals())

    def _add_fragments_to_reaction(self, rxn):
        calc_fragment_volumes(
                self.fragments,
                target_pmol=self.target_pmol_per_frag,
                min_pmol=self.min_pmol_per_frag,
                max_vol_uL=rxn.free_volume.value,
                excess_insert=self.excess_insert,
        )

        for i, frag in enumerate(self.fragments):
            rxn[frag.name].volume = frag.vol_uL, 'µL'
            rxn[frag.name].stock_conc = frag.conc.value, frag.conc.unit
            rxn[frag.name].master_mix = self._is_frag_in_master_mix(i)
            rxn[frag.name].order = 1

        rxn.hold_ratios.volume = self.volume_uL, 'µL'
        rxn.num_reactions = self.num_reactions
        rxn.extra_min_volume = '0.5 µL'

        return rxn

    def _is_frag_in_master_mix(self, i):
        if i == 0:
            return 'bb' in self.master_mix
        else:
            return 'ins' in self.master_mix or str(i) in self.master_mix

@dataclass
class Fragment:
    name: str
    conc_nM: float
    vol_uL: float = None

@dataclass
class Concentration:
    value: float
    unit: str

def parse_fragments(frag_strs):
    """
    Parse fragments from colon-separated strings, i.e. that could 
    be specified on the command-line.

    See the usage text for a description of the syntax of this string.

    Note that if only two fields are specified, they could refer to a name and 
    a concentration, or a concentration and a size.  This is resolved by 
    attempting to interpret each field as a concentration.  Note that this will 
    break if given a name that could be interpreted as a concentration, so 
    don't do that.
    """

    fragments = []

    if len(frag_strs) < 2:
        raise UsageError("must specify at least two fragments")

    for i, frag_str in enumerate(frag_strs):
        fields = frag_str.split(':')
        frag_name = default_fragment_name(i)
        frag_size = None

        if len(fields) == 3:
            frag_name = fields[0]
            frag_conc = conc_from_str(fields[1])
            frag_size = int(fields[2])

        elif len(fields) == 2:
            # Is this (name, conc) or (conc, size)?
            try:
                frag_conc = conc_from_str(fields[0])
                frag_size = int(fields[1])

            except ValueError:
                frag_name = fields[0]
                frag_conc = conc_from_str(fields[1])

        elif len(fields) == 1:
            try:
                frag_conc = conc_from_str(fields[0])

            except ValueError:
                frag_name = fields[0]
                frag_conc = conc_from_str('PCR')

        else:
            raise UsageError("cannot parse fragment '{fragment_str}'")

        if frag_conc.unit == 'ng/µL' and frag_size is None:
            raise UsageError(f"'{frag_str}' specifies a concentration in ng/µL, so the size of the fragment must also be specified (e.g. '{frag_str}:<size>')")

        frag_nM = nM_from_conc(frag_conc, frag_size)
        frag = Fragment(frag_name, frag_nM)
        frag.conc = frag_conc
        fragments.append(frag)

    return fragments

def calc_fragment_volumes(
        frags,
        target_pmol,
        min_pmol,
        max_vol_uL,
        excess_insert=1,
):
    import numpy as np

    # Calculate the ideal amount of each fragment.
    for i, frag in enumerate(frags):
        excess = 1 if i == 0 else excess_insert
        frag.vol_uL = uL_from_pmol(excess * target_pmol, frag.conc_nM)

    # Make sure the maximum volume is not exceeded.
    total_vol_uL = sum(x.vol_uL for x in frags)
    if max_vol_uL < total_vol_uL:
        k = max_vol_uL / total_vol_uL
        for frag in frags:
            frag.vol_uL *= k

    # Warn if any fragment is below the minimum.
    for frag in frags:
        best_pmol = pmol_from_uL(frag.vol_uL, frag.conc_nM)
        if best_pmol < min_pmol:
            warn(f"using {best_pmol:.3f} pmol of {frag.name}, {min_pmol:.3f} pmol recommended.")

def default_fragment_name(i):
    return "backbone" if i == 0 else f"insert #{i}"

def conc_from_str(x):
    import re

    if x.upper().strip() == 'PCR':
        return Concentration(50, 'ng/µL')

    value_pattern = '[0-9.]+'
    unit_pattern = 'ng/[uµ]L|[muµnpf]M'
    conc_pattern = rf'({value_pattern})(\s*({unit_pattern}))?$'

    match = re.match(conc_pattern, x.strip())
    if not match:
        raise ValueError(f"could not interpret '{x}' as a concentration.")

    value, _, unit = match.groups()
    unit = (unit or 'ng/µL').replace('u', 'µ')
    return Concentration(float(value), unit)

def nM_from_conc(conc, num_bp):
    multiplier = {
            'mM': 1e9 / 1e3,
            'µM': 1e9 / 1e6,
            'nM': 1e9 / 1e9,
            'pM': 1e9 / 1e12,
            'fM': 1e9 / 1e15,
    }

    # https://www.neb.com/tools-and-resources/usage-guidelines/nucleic-acid-data
    if num_bp:
        multiplier['ng/µL'] = 1e6 / (650 * num_bp)

    try:
        return conc.value * multiplier[conc.unit]
    except KeyError:
        raise ValueError(f"cannot convert '{conc.unit}' to nM.")

def uL_from_pmol(pmol, conc_nM):
    return 1e3 * pmol / conc_nM

def pmol_from_uL(uL, conc_nM):
    return uL * conc_nM / 1e3


def test_conc_from_str():
    from pytest import approx, raises
    c = Concentration

    assert conc_from_str('1') == c(1.0, 'ng/µL')
    assert conc_from_str('10') == c(10.0, 'ng/µL')
    assert conc_from_str('1.0') == c(1.0, 'ng/µL')
    assert conc_from_str('1 ng/uL') == c(1.0, 'ng/µL')
    assert conc_from_str('1 ng/µL') == c(1.0, 'ng/µL')
    assert conc_from_str('1 fM') == c(1.0, 'fM')
    assert conc_from_str('1 pM') == c(1.0, 'pM')
    assert conc_from_str('1 nM') == c(1.0, 'nM')
    assert conc_from_str('1 uM') == c(1.0, 'µM')
    assert conc_from_str('1 µM') == c(1.0, 'µM')

    assert conc_from_str(' 1 ') == c(1.0, 'ng/µL')
    assert conc_from_str(' 1 nM') == c(1.0, 'nM')
    assert conc_from_str('1 nM ') == c(1.0, 'nM')
    assert conc_from_str(' 1  nM ') == c(1.0, 'nM')

    assert conc_from_str('PCR') == c(50.0, 'ng/µL')
    assert conc_from_str('pcr') == c(50.0, 'ng/µL')

    with raises(ValueError, match="''"):
        conc_from_str('')
    with raises(ValueError, match='xxx'):
        conc_from_str('1 xxx')

def test_nM_from_conc():
    from pytest import approx, raises
    c = Concentration

    assert nM_from_conc(c(1, 'mM'), None) == approx(1e6)
    assert nM_from_conc(c(1, 'µM'), None) == approx(1e3)
    assert nM_from_conc(c(1, 'nM'), None) == approx(1e0)
    assert nM_from_conc(c(1, 'pM'), None) == approx(1e-3)
    assert nM_from_conc(c(1, 'fM'), None) == approx(1e-6)

    assert nM_from_conc(c(1, 'ng/µL'), 100) == approx(1e6/(650 * 100))
    assert nM_from_conc(c(1, 'ng/µL'), 1000) == approx(1e6/(650 * 1000))

def test_fragments_from_strs():
    from pytest import approx, raises
    f = Fragment

    ## 0 fragments
    with raises(UsageError):
        fragments_from_strs('')

    ## 1 fragment
    with raises(UsageError):
        fragments_from_strs(['30nM'])

    ## 2 fragments
     # 3 arguments
    assert fragments_from_strs(['30nM', 'Gene:60:1000']) == [
            f('Backbone', 30),
            f('Gene', approx((60 * 1e6) / (650 * 1000))),
    ]
     # 2 arguments: name, conc
    assert fragments_from_strs(['30nM', 'Gene:60nM']) == [
            f('Backbone', 30),
            f('Gene', 60),
    ]
    with raises(UsageError):
        fragments_from_strs(['30nM', 'Gene:60'])

     # 2 arguments: conc, size
    assert fragments_from_strs(['30nM', '60:1000']) == [
            f('Backbone', 30.0),
            f('Insert #1', approx((60 * 1e6) / (650 * 1000))),
    ]
     # 1 argument: conc
    assert fragments_from_strs(['30nM', '60nM']) == [
            f('Backbone', 30),
            f('Insert #1', 60),
    ]
    with raises(UsageError):
        fragments_from_strs('30nM:60')

    ## 3 fragments
    assert fragments_from_strs(['30nM', '60nM', '61nM']) == [
            f('Backbone', 30),
            f('Insert #1', 60),
            f('Insert #2', 61),
    ]




#!/usr/bin/env python3

import parametrize_from_file
from schema_helpers import *

@parametrize_from_file(schema=app_expected_protocol)
def test_protocol(app, expected):
    actual = app.protocol.format_text()
    print(actual)
    for x in expected:
        assert x in actual





#!/usr/bin/env python3

import stepwise, appcli, autoprop
from inform import fatal, warn, plural
from appcli import Key, DocoptConfig
from stepwise import StepwiseConfig, PresetConfig, pl, ul
from stepwise_mol_bio import Main, require_reagent, merge_dicts
from copy import deepcopy
from operator import not_

class Po4Config(appcli.Config):
    # This should be a general class---provided by `po4`---with the following 
    # features:
    # - Don't load the database until needed.
    # - Will query any key given in the PO₄ database.
    # - Some option to specify the aggregation function for each key.  Can 
    #   probably default to `unanimous()`.
    # - Always yield layer if PO₄ is installed, even if template names aren't 
    #   valid tags.  This leads to better error messages.
    autoload = False

    def load(self, obj):
        try:
            import po4
        except ImportError:
            return

        db = po4.load_db()

        try:
            constructs = [db[x] for x in obj.templates]
        except po4.QueryError:
            return

        values = {}
        
        try:
            values['conc_nM'] = po4.unanimous(
                    x.conc_nM for x in constructs
            )
        except (ValueError, po4.ParseError, po4.QueryError):
            pass

        try:
            values['is_rna'] = po4.unanimous(
                    x.is_rna for x in constructs
            )
        except (ValueError, po4.ParseError, po4.QueryError):
            pass

        yield appcli.Layer(values=values, location=db.name)

@autoprop
class InVitroTranslation(Main):
    """\
Express proteins from linear DNA templates using NEBExpress.

Usage:
    ivtt <templates>... [-p <name>] [-v <µL>] [-n <rxns>] [-x <percent>] 
        [-c <nM>] [-C <nM>] [-mrIX] [-a <name;conc;vol;mm>]... [-t <time>]
        [-T <°C>]

Arguments:
    <templates>
        The templates to express.  The number of reactions will be inferred 
        from this list.

<%! from stepwise_mol_bio import hanging_indent %>\
Options:
    -p --preset <name>
        What default reaction parameters to use.  The following parameters are 
        currently available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    -v --volume <µL>
        The volume of the reaction in µL.  By default, the volume specified by 
        the reaction table in the chosen preset will be used.

    -n --num-reactions <int>
        The number of reactions to set up.  By default, this is inferred from
        the number of templates.

    -x --extra-percent <percent>
        How much extra master mix to prepare, as a percentage of the minimum 
        required  master mix volume.

    -c --template-conc <nM>
        The desired final concentration of template in the reaction.  

    -C --template-stock <nM>
        The stock concentration of the template DNA or mRNA, in units of nM.  
        If not specified, a concentration will be queried from the PO₄ 
        database.  In this case, all templates must be in the database and must 
        have identical concentrations.

    -m --master-mix
        Include the template in the master mix.

    -r --mrna
        Use mRNA as the template instead of DNA.

    -X --no-template
        Don't include the template in the reaction, e.g. as a negative control.

    -I --no-inhibitor
        Don't include RNase inhibitor in the reaction.

    -a --additive <name;conc;vol;mm>
        Add an additional reagent to the reaction.  See `sw reaction -h` for a 
        complete description of the syntax.  This option can be specified 
        multiple times.

    -t --incubation-time <time>         [default: ${app.incubation_time}]
        The amount of time to incubate the reactions.  No unit is assumed, so 
        be sure to include one.  If '0', the incubation step will be removed 
        from the protocol (e.g. so it can be added back at a later point).

    -T --incubation-temperature <°C>    [default: ${app.incubation_temp_C}]
        The temperature to incubate the reactions at, in °C.
"""
    __config__ = [
            Po4Config(),
            DocoptConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.ivtt'),
    ]
    preset_briefs = appcli.config_attr()
    preset_brief_template = '{kit}'

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(StepwiseConfig, 'default_preset'),
    )
    base_reaction = appcli.param(
            Key(PresetConfig, 'reaction'),
            cast=stepwise.MasterMix.from_text,
    )
    title = appcli.param(
            Key(PresetConfig, 'title'),
            Key(PresetConfig, 'kit'),
    )
    templates = appcli.param(
            Key(DocoptConfig, '<templates>'),
    )
    volume_uL = appcli.param(
            Key(DocoptConfig, '--volume', cast=eval),
            default=None,
    )
    default_volume_uL = appcli.param(
            # The difference between `default_volume_uL` and `volume_uL` is 
            # that the default additives are applied to the reaction after the 
            # default volume is set, but before the non-default volume is set.  
            # This allows the volume of the additive to be scaled 
            # proportionally to the volume of the reaction that the additive 
            # was specified for.
            Key(PresetConfig, 'volume_uL'),
            Key(StepwiseConfig, 'default_volume_uL'),
            default=None,
    )
    num_reactions = appcli.param(
            Key(DocoptConfig, '--num-reactions', cast=eval),
            default=None,
            get=lambda self, x: x or len(self.templates),
    )
    extra_percent = appcli.param(
            Key(DocoptConfig, '--extra-percent', cast=float),
            default=10,
    )
    template_conc_nM = appcli.param(
            Key(DocoptConfig, '--template-conc', cast=float),
            Key(PresetConfig, 'template_conc_nM'),
            default=None,
    )
    template_stock_nM = appcli.param(
            Key(DocoptConfig, '--template-stock', cast=float),
            Key(Po4Config, 'conc_nM'),
            Key(PresetConfig, 'template_stock_nM'),
    )
    master_mix = appcli.param(
            Key(DocoptConfig, '--master-mix'),
            default=False,
    )
    use_mrna = appcli.param(
            Key(DocoptConfig, '--mrna'),
            Key(Po4Config, 'is_rna'),
            default=False,
    )
    use_template = appcli.param(
            Key(DocoptConfig, '--no-template', cast=not_),
            default=True,
    )
    use_rnase_inhibitor = appcli.param(
            Key(DocoptConfig, '--no-inhibitor', cast=not_),
            default=True,
    )
    additives = appcli.param(
            Key(DocoptConfig, '--additive'),
            default_factory=list,
    )
    default_additives = appcli.param(
            Key(PresetConfig, 'additives'),
            default_factory=list,
    )
    setup_instructions = appcli.param(
            Key(PresetConfig, 'setup_instructions'),
            default_factory=list,
    )
    setup_footnote = appcli.param(
            Key(PresetConfig, 'setup_footnote'),
            default=None,
    )
    incubation_time = appcli.param(
            Key(DocoptConfig, '--incubation-time'),
            Key(PresetConfig, 'incubation_time'),
    )
    incubation_temp_C = appcli.param(
            Key(DocoptConfig, '--incubation-temp'),
            Key(PresetConfig, 'incubation_temp_C'),
            cast=float,
    )
    incubation_footnote = appcli.param(
            Key(PresetConfig, 'incubation_footnote'),
            default=None,
    )

    def get_protocol(self):
        p = stepwise.Protocol()
        rxn = self.reaction

        p += pl(
                f"Setup {plural(self.num_reactions):# {self.title} reaction/s}{p.add_footnotes(self.setup_footnote)}:",
                rxn,
                ul(*self.setup_instructions),
        )
        if self.incubation_time != '0':
            p += f"Incubate at {self.incubation_temp_C:g}°C for {self.incubation_time}{p.add_footnotes(self.incubation_footnote)}."

        return p

    def get_reaction(self):

        def add_reagents(additives):
            nonlocal i
            # It would be better if there was a utility in stepwise for parsing 
            # `sw reaction`-style strings.  Maybe `Reagent.from_text()`.
            for i, additive in enumerate(additives, i):
                reagent, stock_conc, volume, master_mix = additive.split(';')
                rxn[reagent].stock_conc = stock_conc
                rxn[reagent].volume = volume
                rxn[reagent].master_mix = {'+': True, '-': False, '': False}[master_mix.strip()]
                rxn[reagent].order = i

        rxn = deepcopy(self.base_reaction)
        rxn.num_reactions = self.num_reactions

        for i, reagent in enumerate(rxn):
            reagent.order = i

        if self.default_volume_uL:
            rxn.hold_ratios.volume = self.default_volume_uL, 'µL'

        add_reagents(self.default_additives)

        if self.volume_uL:
            rxn.hold_ratios.volume = self.volume_uL, 'µL'

        add_reagents(self.additives)

        if self.use_mrna:
            template = 'mRNA'
            require_reagent(rxn, 'mRNA')
            del_reagent_if_present(rxn, 'DNA')
            del_reagents_by_flag(rxn, 'dna')
        else:
            template = 'DNA'
            require_reagent(rxn, 'DNA')
            del_reagent_if_present(rxn, 'mRNA')
            del_reagents_by_flag(rxn, 'mrna')

        rxn[template].name = f"{','.join(self.templates)}"
        rxn[template].master_mix = self.master_mix
        rxn[template].hold_conc.stock_conc = self.template_stock_nM, 'nM'

        if self.template_conc_nM:
            rxn[template].hold_stock_conc.conc = self.template_conc_nM, 'nM'
        elif self.use_template:
            warn("Template concentrations must be empirically optimized.\nThe default value is just a plausible starting point.")

        if not self.use_template:
            del rxn[template]

        if not self.use_rnase_inhibitor:
            del_reagents_by_flag(rxn, 'rnase')

        # Make sure the template is added last.
        rxn[template].order = i+1

        if self.use_template:
            rxn.fix_volumes(template)

        rxn.extra_percent = self.extra_percent

        return rxn

def del_reagent_if_present(rxn, key):
    if key in rxn:
        del rxn[key]

def del_reagents_by_flag(rxn, flag):
    reagents = list(rxn.iter_reagents_by_flag(flag))
    for reagent in reagents:
        del rxn[reagent.key]


if __name__ == '__main__':
    InVitroTranslation.main()



#!/usr/bin/env python3

import parametrize_from_file
from stepwise_mol_bio._utils import *
from schema_helpers import *

@parametrize_from_file(
        schema=Schema({
            'given': eval,
            'length': eval,
            **error_or({
                'expected': eval,
            }),
        }),
)
def test_match_len(given, length, expected, error):
    with error:
        assert match_len(given, length) == expected




#!/usr/bin/env python3
# vim: tw=50

import stepwise, appcli, autoprop
from inform import plural
from stepwise_mol_bio import Assembly

@autoprop
class Ligate(Assembly):
    """\
Assemble restrictions digested DNA fragments using T4 DNA ligase.

Usage:
    ligate <backbone> <inserts>... [-n <int>] [-v <µL>] [-k] [options]

Arguments:
{FRAGMENT_DOC}

Options:
{OPTION_DOC}

    -k --kinase
        Add T4 polynucleotide kinase (PNK) to the reaction.  This is necessary 
        to ligate ends that are not already 5' phosphorylated (e.g. annealed 
        oligos, PCR products).
"""

    excess_insert = appcli.param(
            '--excess-insert',
            cast=float,
            default=3,
    )
    use_kinase = appcli.param(
            '--kinase',
            default=False,
    )

    def get_reaction(self):
        rxn = stepwise.MasterMix.from_text('''\
        Reagent               Stock        Volume  Master Mix
        ================  =========   ===========  ==========
        water                         to 20.00 μL         yes
        T4 ligase buffer        10x       2.00 μL         yes
        T4 DNA ligase      400 U/μL       1.00 μL         yes
        T4 PNK              10 U/μL       1.00 μL         yes
        ''')
        if not self.use_kinase:
            del rxn['T4 PNK']

        return self._add_fragments_to_reaction(rxn)

    def get_protocol(self):
        p= stepwise.Protocol()
        p+= f"""\
Run {plural(self.num_reactions):# ligation reaction/s} [1]:

{self.reaction}
"""
        p += """\
Incubate at the following temperatures:

- 25°C for 15 min
- 65°C for 10 min
"""
        p += """\
Transform 2 µL.
"""
        p.footnotes[1] = """\
https://preview.tinyurl.com/y7gxfv5m
"""
        return p

if __name__ == '__main__':
    Ligate.main()



#!/usr/bin/env python3

import stepwise
import appcli
import autoprop
from inform import plural
from stepwise_mol_bio import Assembly

@autoprop
class Gibson(Assembly):
    """\
Perform a Gibson assembly reaction, using the NEB master mix (E2611).

Usage:
    gibson_assembly <backbone> <inserts>... [options]

Arguments:
{FRAGMENT_DOC}

Options:
{OPTION_DOC}
"""

    def get_reaction(self):
        rxn = stepwise.MasterMix()
        rxn.volume = '20 µL'

        rxn['Gibson master mix (NEB E2611)'].volume = rxn.volume / 2
        rxn['Gibson master mix (NEB E2611)'].stock_conc = "2x"
        rxn['Gibson master mix (NEB E2611)'].master_mix = True
        rxn['Gibson master mix (NEB E2611)'].order = 2

        return self._add_fragments_to_reaction(rxn)

    def get_protocol(self):
        p = stepwise.Protocol()
        n = self.num_reactions
        incubation_time = '15 min' if len(self.fragments) <= 3 else '1h'

        p += f"""\
Setup the Gibson assembly {plural(n):reaction/s} [1]:

{self.reaction}
"""
        p += f"""\
Incubate at 50°C for {incubation_time}.
"""
        p += f"""\
Transform 2 µL.
"""
        p.footnotes[1] = """\
https://preview.tinyurl.com/ychbvkra
"""
        return p

    def get_usage(self):
        return format_docstring(self, self.__doc__)

if __name__ == '__main__':
    Gibson.main()

# vim: tw=50



#!/usr/bin/env python3

import stepwise
import autoprop
import appcli

from stepwise import (
        PresetConfig, StepwiseConfig, Quantity, pl, ul, oxford_comma,
)
from appcli import Key, Method, DocoptConfig
from freezerbox import (
        ReagentConfig, MakerArgsConfig, ParseError, unanimous, 
        convert_conc_unit, group_by_identity, join_lists,
        parse_volume_uL as parse_strict_volume_uL,
)
from stepwise_mol_bio import (
        Cleanup, Gel, SpinCleanup, ConfigError, merge_dicts, comma_list,
)
from operator import not_
from more_itertools import one
from inform import plural
from math import ceil, inf

def parse_conc(x):
    return Quantity.from_string_or_float(x, 'µg/µL')

def parse_volume_uL(x):
    return float(x)

def parse_samples(values):
    return [Sample.from_colon_separated_string(x) for x in values]

class Sample:
    __config__ = [
            ReagentConfig(
                db_getter=lambda self: self.app.db,
                tag_getter=lambda self: self.name,
                transform=one,
            )
    ]

    def _calc_mass_ug(self):
        try:
            stock_conc_ug_uL = convert_conc_unit(self.stock_conc, self.mw, 'µg/µL')
        except ParseError:
            err = ConfigError(sample=self)
            err.brief = "can't calculate mass in µg for sample: {sample.name!r}"
            err.info += "MW: {sample.mw}"
            err.info += "stock conc: {sample.stock_conc}"
            err.info += "volume: {sample.stock_conc}"
            raise err
        else:
            return Quantity(stock_conc_ug_uL.value * self.volume_uL, 'µg')

    name = appcli.param()
    molecule = appcli.param(
            Method(lambda self: self.app.default_molecule),
            Key(ReagentConfig, 'molecule'),
            default='RNA',
            ignore=None,
    )
    stock_conc = appcli.param(
            Method(lambda self: self.app.default_stock_conc),
            Key(ReagentConfig, 'conc'),
            ignore=None,
    )
    volume_uL = appcli.param(
            Method(lambda self: self.app.default_volume_uL),
            ignore=None,
    )
    mw = appcli.param(
            Key(ReagentConfig, 'mw'),
            default=None,
            ignore=None,
    )
    mass_ug = appcli.param(
            Method(_calc_mass_ug),
            ignore=None,
    )

    @classmethod
    def from_colon_separated_string(cls, value):
        fields = {i: x for i, x in enumerate(value.split(':'))}

        name = fields[0]
        molecule = fields.get(1) or None
        stock_conc = fields.get(2) or None
        volume_uL = fields.get(3) or None

        if stock_conc:
            stock_conc = parse_conc(stock_conc)
        if volume_uL:
            volume_uL = parse_volume_uL(volume_uL)

        return cls(
                name=name,
                molecule=molecule,
                stock_conc=stock_conc,
                volume_uL=volume_uL,
        )

    def __init__(self, name, molecule=None, stock_conc=None, volume_uL=None, mass_ug=None):
        self.app = None
        self.name = name
        self.molecule = molecule
        self.stock_conc = stock_conc
        self.volume_uL = volume_uL
        self.mass_ug = mass_ug

    def __repr__(self):
        attrs = ['name', 'molecule', 'stock_conc', 'volume_uL', 'mass_ug']
        attr_reprs = [
                f'{k}={v!r}'
                for k in attrs
                if (v := getattr(self, k, None))
        ]
        return f'Sample({", ".join(attr_reprs)})'

    def bind(self, app, force=False):
        if not self.app or force:
            self.app = app

@autoprop
class PagePurify(Cleanup):
    """\
Purify nucleic acids by PAGE.

Usage:
    page_purify <samples>... [-p <preset>] [-P <gel>] [-c <conc>] [-v <µL>]
        [-b <bands>] [-R | -D] [-C]

Arguments:
    <samples>
        A description of each crude sample to purify.  Each argument can 
        contain several pieces of information, separated by colons as follows:

            name[:molecule[:conc[:volume]]]

        name:
            The name of the product.  If this is corresponds to a tag in the 
            FreezerBox database, default values for the other parameters will 
            be read from the database.

        molecule:
            What kind of molecule the product is: either "RNA" or "DNA" 
            (case-insensitive).  Use `--rna` or `--dna` if all samples are the 
            same molecule.

        conc:
            The concentration of the product.  This may include a unit.  If no 
            unit is specified, µg/µL is assumed.  Use `--conc` if all samples 
            are the same concentration.

        volume:
            The volume of the product to load on the gel, in µL.  Do not 
            include a unit.  Use `--volume` if all samples have the same 
            volume.
            
<%! from stepwise_mol_bio import hanging_indent %>\
Options:
    -p --preset <name>          [default: ${app.preset}]
        The default parameters to use.  The following presets are available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    -P --gel-preset <name>
        The default gel electrophoresis parameters to use.  See `sw gel -h` for 
        a list of available presets.

    -c --conc <float>
        The concentration of each sample.  This may include a unit.  If no unit 
        is specified, µg/µL is assumed.  This is superseded by concentrations 
        specified in the <samples> argument, but supersedes concentrations 
        found in the FreezerBox database.

    -v --volume <µL>
        The volume of each sample to load on the gel, in µL.  Do not include a 
        unit.  This is superseded by volumes specified in the <samples> 
        argument.

    -b --bands <names>
        A comma separated list of names identifying the bands to cut out of the 
        gel.  Typically these would be the lengths of the desired products, 
        e.g. "400 nt" or "1.5 kb".

    -R --rna
        Assume that each sample is RNA.  This is superseded by the <samples> 
        argument, but supersedes the FreezerBox database.

    -D --dna
        Assume that each sample is DNA.  This is superseded by the <samples> 
        argument, but supersedes the FreezerBox database.

    -C --no-cleanup
        Don't include the final spin-column purification step in the protocol.

Configuration:
    Default values for this protocol can be specified in any of the following 
    stepwise configuration files:

        ${hanging_indent(app.config_paths, 8)}

    molbio.page_purify.default_preset:
        The default value for the `--preset` option.

    molbio.page_purify.presets:
        Named groups of default reaction parameters.  Typically each preset 
        corresponds to a particular kit or protocol.  See below for the various 
        settings that can be specified in each preset.

    molbio.page_purify.presets.<name>.conc
        The default value for the `--conc` option.  Note that if this option is 
        set, concentrations will never be read from the FreezerBox database.

    molbio.page_purify.presets.<name>.volume_uL
        The default value for the `--volume` option.

    molbio.page_purify.presets.<name>.molecule
        The default value for the `--rna`/`--dna` options.  This should either 
        be "RNA" or "DNA".

    molbio.page_purify.presets.<name>.gel_preset
        The default value for the `--gel-preset` option.

    molbio.page_purify.presets.<name>.bands
        The default value for the `--bands` option.

    molbio.page_purify.presets.<name>.cleanup_preset
        The default presets to use for the spin column cleanup step after 
        recovering the DNA/RNA from the gel.  See `sw spin_cleanup -h` for a 
        list of valid presets.  This option should be a dictionary with the 
        keys 'rna' and 'dna', each specifying the preset to use with samples of 
        the corresponding type.  Alternatively, this option can be 'false' to 
        indicate that the cleanup step should be skipped (see `--no-cleanup`).

Database:
    PAGE purification protocols can appear in the "Cleanups" column of a 
    FreezerBox database:

        page-purify [preset=<name>] [gel=<name>] [conc=<conc>] [volume=<µL>]
    
    preset=<name>
        See `--preset`.

    gel=<name>
        See `--gel-preset`.

    conc=<conc>
        See `--conc`.  Must include a unit.

    volume=<µL>
        See `--volume`.  Must include a unit.
"""
    __config__ = [
            DocoptConfig(),
            MakerArgsConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.page_purify'),
    ]
    Sample = Sample
    preset_briefs = appcli.config_attr()
    config_paths = appcli.config_attr()

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(MakerArgsConfig, 'preset'),
            Key(StepwiseConfig, 'default_preset'),
    )
    samples = appcli.param(
            Key(DocoptConfig, '<samples>', cast=parse_samples),
            Method(lambda self: [Sample(name=x.tag) for x in self.products]),
    )
    default_stock_conc = appcli.param(
            Key(DocoptConfig, '--conc', cast=parse_conc),
            Key(MakerArgsConfig, 'conc', cast=Quantity.from_string),
            Key(PresetConfig, 'conc', cast=Quantity.from_string),
    )
    default_volume_uL = appcli.param(
            Key(DocoptConfig, '--volume', cast=parse_volume_uL),
            Key(MakerArgsConfig, 'volume', cast=parse_strict_volume_uL),
            Key(PresetConfig, 'volume_uL', cast=parse_volume_uL),
            default=5,
    )
    default_molecule = appcli.param(
            Key(DocoptConfig, '--dna', cast=lambda x: 'DNA'),
            Key(DocoptConfig, '--rna', cast=lambda x: 'RNA'),
            Key(PresetConfig, 'molecule'),
    )
    gel_preset = appcli.param(
            Key(DocoptConfig, '--gel-preset'),
            Key(MakerArgsConfig, 'gel'),
            Key(PresetConfig, 'gel_preset'),
    )
    gel_percent = appcli.param(
            Key(DocoptConfig, '--gel-percent'),
            default=None,
    )
    gel_run_volts = appcli.param(
            Key(DocoptConfig, '--gel-run-volts'),
            default=None,
    )
    gel_run_time_min = appcli.param(
            Key(DocoptConfig, '--gel-run-time-min'),
            default=None,
    )
    desired_bands = appcli.param(
            Key(DocoptConfig, '--bands', cast=comma_list),
            Key(PresetConfig, 'bands'),
            default_factory=list,
    )
    rna_cleanup_preset = appcli.param(
            Key(PresetConfig, 'cleanup_preset.rna'),
    )
    dna_cleanup_preset = appcli.param(
            Key(PresetConfig, 'cleanup_preset.dna'),
    )
    cleanup = appcli.param(
            Key(DocoptConfig, '--no-cleanup', cast=not_),
            Key(PresetConfig, 'cleanup_preset', cast=bool),
    )

    group_by = {
            'preset': group_by_identity,
            'gel_preset': group_by_identity,
    }
    merge_by = {
            'samples': join_lists,
    }

    def __init__(self, samples):
        self.samples = samples

    def get_protocol(self):
        self._bind_samples()

        p = stepwise.Protocol()
        p += self.gel_electrophoresis_steps
        p += self.gel_extraction_steps
        p += self.product_recovery_steps
        return p

    def get_gel_electrophoresis_steps(self):
        p = stepwise.Protocol()
        n = plural(self.samples)
        gel = Gel(self.gel_preset)

        mix = gel.sample_mix
        k = mix.volume / mix['sample'].volume

        for sample in self.samples:
            sample.stock_conc
            sample.sample_volume_per_lane_uL = min(
                    sample.volume_uL,
                    sample.volume_uL / ceil(sample.mass_ug / '20 µg'),
                    mix['sample'].volume.value or inf,
            )
            sample.load_volume_per_lane_uL = k * sample.sample_volume_per_lane_uL
            sample.num_lanes = int(ceil(sample.volume_uL / sample.sample_volume_per_lane_uL))

        conc_vol_groups = group_by_identity(
                self.samples,
                key=lambda x: (x.stock_conc, x.volume_uL),
        )
        for (conc, volume_uL), group in conc_vol_groups:
            prep_gel = Gel(self.gel_preset)
            prep_gel.num_samples = len(group)
            mix = prep_gel.sample_mix

            mix.hold_ratios.volume = k * volume_uL, 'µL'
            mix['sample'].name = ','.join(str(x.name) for x in group)
            mix['sample'].stock_conc = conc

            p += prep_gel.prep_step

        if x := self.gel_percent:
            gel.gel_percent = x
        if x := self.gel_run_volts:
            gel.run_volts = x
        if x := self.gel_run_time_min:
            gel.run_time_min = x

        try:
            gel.load_volume_uL = unanimous(
                    x.load_volume_per_lane_uL
                    for x in self.samples
            )
        except ValueError:
            gel.load_volume_uL = 0

            p += pl(f"Load the {n:sample/s} in the gel as follows:", u := ul())
            for sample in self.samples:
                num_lanes = f" in each of {m} lanes" if (m := sample.num_lanes) != 1 else ""
                u += f"{sample.load_volume_per_lane_uL:.2g} µL {sample.name}{num_lanes}"

        p += gel.run_step

        return p

    def get_gel_extraction_steps(self):
        p = stepwise.Protocol()
        names = self._cluster_names_by_molecule()

        desired = oxford_comma(x) if (x := self.desired_bands) else 'desired'
        n = plural(max(len(self.desired_bands), len(self.samples)))
        f = "Based on Fitzy's DNA PAGE purification protocol, [Nilson2013], and [Petrov2013]."

        p += pl(f"Cut the {desired} {n:band/s} out of the gel{p.add_footnotes(f)}.", ul(
            "Place the gel over a TLC plate.",
            "Use a UV light to visualize the RNA (dark spot).",
            "Consider visualizing remaining gel to ensure that all desired RNA was excised.",
        ))
        p += pl(f"Crush the gel {n:slice/s}.", ul(
            "Poke a hole in the bottom of a 0.65 mL tube with a 27 g needle.",
            "Place gel slice inside the 0.65 mL tube.",
            "Place the 0.65 mL tube inside a 1.5 mL tube.",
            "Spin at max speed for 5 min.",
        ))

        eb = elution_buffer()
        eb.hold_ratios.volume = 500 * len(self.samples), 'µL'
        p += pl(
                "Resuspend gel in 400 µL PAGE elution buffer:",
                eb,
        )

        if names['RNA']:
            p += f"Incubate {join_if(names['RNA'], names['DNA'])}at 4°C overnight with end-over-end mixing."

        if names['DNA']:
            p += f"Incubate {join_if(names['DNA'], names['RNA'])}at 55°C overnight with 800 rpm mixing."

        return p
            
    def get_product_recovery_steps(self):
        p = stepwise.Protocol()
        names = self._cluster_names_by_molecule()

        # Filter material:
        # - Corning has a good guide on which material to select:
        #   https://www.corning.com/catalog/cls/documents/selection-guides/t_filterselectionguide.pdf
        #
        # - I want 0.22 µm, because that's the standard size for filter 
        #   sterilizing biological buffers (that's not my application here, but 
        #   I can see myself wanting to do that).
        #
        # - I want cellulose acetate filters.  Nylon and cellulose nitrate have 
        #   high DNA binding, which will cause me to lose material.  The 
        #   downside to cellulose acetate is that it has a wetting agent that 
        #   will end up in the sample.  However, this will be removed by the 
        #   Zymo column in the subsequent step.
        #
        # - Product number: 8161 (non sterile)
        #
        # Centrifugation speed:
        # - Fitzy's DNA PAGE purification protocol calls for 4 min at 7000 rpm
        # - The Corning guide (see above) includes an agarose gel purification 
        #   protocol, which calls for 13,000g for 5-20 min.  But this protocol 
        #   has no incubation step, so I gather that the spin is supposed to 
        #   pull the solvent out of the gel.  I probably don't need to go so 
        #   fast, but why not go as fast as the columns can tolerate?
        f = "Nylon and cellulose nitrate have high DNA-binding: https://tinyurl.com/3pkyc8dr"
        p += pl(
                "Remove gel debris by spin filtration:",
                ul(
                    f"Load samples onto a 0.22 µm cellose-acetate Spin-X column{p.add_footnotes(f)}.",
                    "Spin at 13,000g for 5 min."
                ),
        )

        if self.cleanup:
            if names['RNA']:
                p += cleanup(self.rna_cleanup_preset, names['RNA'], names['DNA'])

            if names['DNA']:
                p += cleanup(self.dna_cleanup_preset, names['DNA'], names['RNA'])

        return p

    def _bind_samples(self):
        assert self.samples

        for sample in self.samples:
            sample.bind(self)

    def _cluster_names_by_molecule(self):
        names = {'RNA': [], 'DNA': []}
        for sample in self.samples:
            names[sample.molecule].append(sample.name)
        return names

    @classmethod
    def _solo_maker_factory(cls, product):
        app = super()._solo_maker_factory(product)
        app._bind_samples()
        return app

def elution_buffer():
    return stepwise.MasterMix("""\
            Reagent               Stock      Volume
            ===================  ======  ==========
            nuclease-free water          to 1000 µL
            Tris, pH 7.5            1 M       10 µL
            NaCl                    5 M      100 µL
            EDTA                 500 mM        2 µL
            SDS                     10%       10 µL
    """)

def join_if(items, cond):
    if not cond:
        return ''
    else:
        return stepwise.oxford_comma(items) + ' '

def cleanup(preset, products, show_products):
    sc = SpinCleanup(preset)
    sc.sample_volume_uL = 400

    if show_products:
        sc.products = products
        sc.show_product_names = True

    return sc.protocol


if __name__ == '__main__':
    PagePurify.main()



#!/usr/bin/env python3

import parametrize_from_file
from stepwise_mol_bio import ivt
from schema_helpers import *

@parametrize_from_file(schema=app_expected_reaction)
def test_reaction(app, expected):
    assert app.reaction.format_text() == expected

@parametrize_from_file(schema=app_expected_reaction)
def test_dnase_reaction(app, expected):
    assert app.dnase_reaction.format_text() == expected

@parametrize_from_file(schema=app_expected_protocol)
def test_protocol(app, expected):
    actual = app.protocol.format_text()
    print(actual)
    for x in expected:
        assert x in actual

@parametrize_from_file(schema=app_expected)
def test_short(app, expected):
    assert app.short == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_pick_by_short(values, is_short, expected):
    assert ivt.pick_by_short(values, is_short) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_affected_by_short(values, expected):
    assert ivt.affected_by_short(values) == expected

@parametrize_from_file(
        schema=Schema({
            'template': str,
            **error_or({
                'expected': str,
            }),
        }),
)
def test_transcribe(template, expected, error):
    with error:
        assert ivt.transcribe(template) == expected

@parametrize_from_file(schema=db_expected)
def test_make(db, expected):
    for tag in expected:
        assert db[tag].seq == expected[tag]['seq']
        assert db[tag].dependencies == expected[tag]['dependencies']
        assert db[tag].molecule == 'RNA'
        assert db[tag].is_single_stranded == True





#!/usr/bin/env python3

from appcli import Key, DocoptConfig
from stepwise import UsageError, StepwiseConfig, PresetConfig
from stepwise_mol_bio import Pcr

class Qpcr(Pcr):
    """\
Amplify a DNA template using polymerase chain reaction (PCR).

Note that this protocol is simply an alias for the PCR protocol, with a 
different default preset.

Usage:
    {pcr}
"""
    __config__ = [
            DocoptConfig(usage_getter=lambda self: self.format_usage()),
            PresetConfig(),
            StepwiseConfig('molbio.qpcr'),
            StepwiseConfig('molbio.pcr'),
    ]
    def format_usage(self):
        return self.__doc__.format(
                pcr=super().__doc__
                        .split('Usage:', 1)[1]
                        .replace('pcr', 'qpcr', 1)
                        .strip(),
        )

if __name__ == '__main__':
    Qpcr.main()







#!/usr/bin/env python3

import stepwise, appcli, autoprop
from inform import warn
from appcli import Key, DocoptConfig
from stepwise import StepwiseConfig, PresetConfig, Quantity, oxford_comma
from stepwise_mol_bio import Cleanup, format_sec, merge_dicts
from freezerbox import MakerArgsConfig, group_by_identity, parse_volume_uL
from more_itertools import always_iterable

def ng_uL(x):
    return Quantity(x, 'ng/µL')

@autoprop
class SpinCleanup(Cleanup):
    """\
Purify a PCR reaction using a silica spin column.

Usage:
    spin_cleanup [<preset>] [-s <µL>] [-d <buffer>] [-v <µL>]

<%! from stepwise_mol_bio import hanging_indent %>\
Arguments:
    <preset>                        [default: ${app.preset}]
        The default parameters to use.  Typically these correspond to 
        commercial kits:

        ${hanging_indent(app.preset_briefs, 8*' ')}

Options:
    -s --sample-volume <µL>
        The volume of the sample, in µL.

    -d --elute-buffer <name>
        The buffer to elute in.

    -v --elute-volume <µL>
        The volume of purified DNA/RNA to elute, in µL.  The default value 
        depends on the preset, but can usually be lowered to get more 
        concentrated product.  A warning will be displayed if the requested 
        volume is lower than the minimum recommended by the kit manufacturer.

Configuration:
    Default values for this protocol can be specified in any of the following 
    stepwise configuration files:

        ${hanging_indent(app.config_paths, 8)}

    molbio.spin_cleanup.default_preset:
        The default value for the `--preset` option.

    molbio.spin_cleanup.presets:
        Named groups of default reaction parameters.  Typically each preset 
        corresponds to a particular kit or protocol.  See below for the various 
        settings that can be specified in each preset.

    molbio.spin_cleanup.presets.<name>.protocol_name
        How to refer to the whole protocol.  Commonly this is the name of the 
        spin column kit.

    molbio.spin_cleanup.presets.<name>.protocol_link
        A link (typically minified) to the complete protocol, e.g. as published 
        by the manufacturer of the columns.  This is not required, but if 
        specified, will be included in the protocol as a footnote.

    molbio.spin_cleanup.presets.<name>.column_name
        How to refer to the specific spin column used in the protocol.

    molbio.spin_cleanup.presets.<name>.spin_speed_g
        How fast to spin the column in each centrifugation step, in units of 
        g-force.

    molbio.spin_cleanup.presets.<name>.column_capacity_ug
        The maximum binding capacity of the column, in µg.  This information is 
        added to the protocol as a footnote.

    molbio.spin_cleanup.presets.<name>.sample_type
        How to generically refer to the sample in the protocol, e.g. "DNA".

    molbio.spin_cleanup.presets.<name>.sample_volume_uL
        The volume of sample to load on the column, in µL.  Alternatively, this 
        can be a dictionary with keys 'min' and/or 'max' specifying the minimum 
        and maximum allowed sample volumes, respectively.

    molbio.spin_cleanup.presets.<name>.bind_buffer
        The name(s) of the buffer(s) to use to bind the sample to column.  This 
        can be either a string or a list of strings.  Use a list to specify 
        that multiple buffers (e.g. binding buffer and ethanol) should be mixed 
        with the sample before it is loaded on the column.  If this option is a 
        list, the `bind_volume_uL` and `bind_volume_x` options must also be 
        lists of the same length (or left unspecified).
        
    molbio.spin_cleanup.presets.<name>.bind_volume_uL
        How much `bind_buffer` to use, in µL.  This can be either a number or a 
        list of numbers; see `bind_buffer` for more details.  This takes 
        precedence over the `bind_volume_x` setting.  

    molbio.spin_cleanup.presets.<name>.bind_volume_x
        How much `bind_buffer` to use, as a multiple of the sample volume.  
        This can be a number or a list of numbers; see `bind_buffer` for more 
        details.  This is superseded by the `bind_volume_uL` setting.  

    molbio.spin_cleanup.presets.<name>.bind_spin_sec
        How long to centrifuge the column during the bind step.

    molbio.spin_cleanup.presets.<name>.bind_vacuum
        Whether or not to use a vacuum manifold for the bind step.  The default 
        is False.  If True, the `bind_spin_sec` option is ignored.

    molbio.spin_cleanup.presets.<name>.pH_buffer
        The name of the buffer to use when adjusting the pH of the sample.

    molbio.spin_cleanup.presets.<name>.pH_volume_uL
        How much `pH_buffer` to use, in µL.  This takes precedence over the 
        `pH_volume_x` setting.

    molbio.spin_cleanup.presets.<name>.pH_volume_x
        How much `pH_buffer` to use, as a multiple of the sample volume.  
        This is superseded by the `pH_volume_uL` setting.

    molbio.spin_cleanup.presets.<name>.pH_color
        The color the sample/binding buffer should be after reaching the 
        correct pH.

    molbio.spin_cleanup.presets.<name>.wash_buffer
        The name of the buffer to use when washing the column.  This can either 
        be a string or a list of strings.  Use a list to specify that there 
        should be multiple wash steps.  If this option is a list, the 
        `wash_volume_uL`, `wash_spin_sec`, and `wash_vacuum` options must also 
        be lists of the same length (or left unspecified).

    molbio.spin_cleanup.presets.<name>.wash_volume_uL
        The volume of `wash_buffer` to use, in µL.  This can either be a number 
        or a list of numbers; see `wash_buffer` for more details.

    molbio.spin_cleanup.presets.<name>.wash_spin_sec
        How long to centrifuge the column during the wash step.  This can 
        either be a number or a list of numbers; see `wash_buffer` for more 
        details.

    molbio.spin_cleanup.presets.<name>.wash_vacuum
        Whether or not to use a vacuum manifold for the wash step.  This can 
        either be a boolean or a list of booleans; see `wash_buffer` for more 
        details.  The default is False.  If True, the `wash_spin_sec` option is 
        ignored.

    molbio.spin_cleanup.presets.<name>.dry_spin_sec
        How long to centrifuge the column after the wash step(s), e.g. to 
        remove any residual ethanol.  If left unspecified, this step will not 
        be included in the protocol.

    molbio.spin_cleanup.presets.<name>.elute_buffer
        The default value for the `--elute-buffer` flag.

    molbio.spin_cleanup.presets.<name>.elute_volume_uL
        The default value for the `--elute-volume` flag.

    molbio.spin_cleanup.presets.<name>.elute_min_volume_uL
        The minimum recommended volume to elute in.  Smaller volumes can still 
        be specified, but will be accompanied by a warning.

    molbio.spin_cleanup.presets.<name>.elute_wait_sec
        How long to incubate the column with elution buffer before eluting, in 
        seconds.

    molbio.spin_cleanup.presets.<name>.elute_spin_sec
        How long to centrifuge the column when eluting.

Database:
    Spin-column cleanup protocols can appear in the "Cleanups" column of a 
    FreezerBox database:

        spin-cleanup [<preset>] [volume=<µL>] [buffer=<name>]
    
    <preset>
        See `<preset>`.

    volume=<µL>
        See `--elute-volume`.  Must specify a unit.

    buffer=<µL>
        See `--elute-buffer`.
"""
    __config__ = [
            DocoptConfig(),
            MakerArgsConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.spin_cleanup'),
    ]
    preset_briefs = appcli.config_attr()
    config_paths = appcli.config_attr()
    preset_brief_template = '{protocol_name}'

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '<preset>'),
            Key(MakerArgsConfig, 1),
            Key(StepwiseConfig, 'default_preset'),
    )
    protocol_name = appcli.param(
            Key(PresetConfig, 'protocol_name'),
    )
    protocol_link = appcli.param(
            Key(PresetConfig, 'protocol_link'),
            default=None,
    )
    column_name = appcli.param(
            Key(PresetConfig, 'column_name'),
            default='silica spin column',
    )
    spin_speed_g = appcli.param(
            Key(PresetConfig, 'spin_speed_g'),
            default=None,
    )
    column_capacity_ug = appcli.param(
            Key(PresetConfig, 'column_capacity_ug'),
            default=None,
    )
    sample_type = appcli.param(
            Key(PresetConfig, 'sample_type'),
            default='DNA',
    )
    sample_volume_uL = appcli.param(
            Key(DocoptConfig, '--sample-volume', cast=float),
            default=None,
    )
    target_sample_volume_uL = appcli.param(
            Key(PresetConfig, 'sample_volume_uL'),
            default=None,
    )
    bind_buffer = appcli.param(
            Key(PresetConfig, 'bind_buffer'),
    )
    bind_volume_uL = appcli.param(
            Key(PresetConfig, 'bind_volume_uL'),
            default=None
    )
    bind_volume_x = appcli.param(
            Key(PresetConfig, 'bind_volume_x'),
            default=None
    )
    bind_spin_sec = appcli.param(
            Key(PresetConfig, 'bind_spin_sec'),
            default=None
    )
    bind_vacuum = appcli.param(
            Key(PresetConfig, 'bind_vacuum'),
            default=False,
    )
    ph_buffer = appcli.param(
            Key(PresetConfig, 'pH_buffer'),
            default=None,
    )
    ph_volume_uL = appcli.param(
            Key(PresetConfig, 'pH_volume_uL'),
            default=None
    )
    ph_volume_x = appcli.param(
            Key(PresetConfig, 'pH_volume_x'),
            default=None
    )
    ph_color = appcli.param(
            Key(PresetConfig, 'pH_color'),
    )
    wash_buffer = appcli.param(
            Key(PresetConfig, 'wash_buffer'),
    )
    wash_volume_uL = appcli.param(
            Key(PresetConfig, 'wash_volume_uL'),
    )
    wash_spin_sec = appcli.param(
            Key(PresetConfig, 'wash_spin_sec'),
            default=None,
    )
    wash_vacuum = appcli.param(
            Key(PresetConfig, 'wash_vacuum'),
            default=False,
    )
    dry_spin_sec = appcli.param(
            Key(PresetConfig, 'dry_spin_sec'),
            default=None,
    )
    elute_buffer = appcli.param(
            Key(DocoptConfig, '--elute-buffer'),
            Key(MakerArgsConfig, 'buffer'),
            Key(PresetConfig, 'elute_buffer'),
    )
    elute_volume_uL = appcli.param(
            Key(DocoptConfig, '--elute-volume', cast=float),
            Key(MakerArgsConfig, 'volume', cast=parse_volume_uL),
            Key(PresetConfig, 'elute_volume_uL'),
    )
    elute_min_volume_uL = appcli.param(
            Key(PresetConfig, 'elute_min_volume_uL'),
            default=None,
    )
    elute_wait_sec = appcli.param(
            Key(PresetConfig, 'elute_wait_sec'),
            default=None,
    )
    elute_spin_sec = appcli.param(
            Key(PresetConfig, 'elute_spin_sec'),
    )

    group_by = {
            'preset': group_by_identity,
            'elute_buffer': group_by_identity,
            'elute_volume_uL': group_by_identity,
    }

    def __init__(self, preset=None):
        if preset is not None:
            self.preset = preset

    def get_protocol(self):
        p = stepwise.Protocol()
        pl = stepwise.paragraph_list()
        ul = stepwise.unordered_list()

        def break_if_too_long(pl, ul, n=4):
            if len(ul) > n:
                ul = stepwise.unordered_list()
                pl += ul
            return ul

        footnotes = []
        if self.protocol_link:
            footnotes.append(self.protocol_link)
        if self.column_capacity_ug:
            footnotes.append(f"Column capacity: {self.column_capacity_ug} µg")

        if self.products and self.show_product_names:
            product_names = oxford_comma(self.products) + ' '
        else:
            product_names = ''

        p += pl
        pl += f"Purify {product_names}using {self.protocol_name}{p.add_footnotes(*footnotes)}:"
        pl += ul

        if self.spin_speed_g:
            ul += f"Perform all spin steps at {self.spin_speed_g}g."

        ## Dilute
        if x := self.target_sample_volume_uL:
            v = self.sample_volume_uL

            if not isinstance(x, dict):
                target = f'{x} µL'
                skip = v and v == x
                self.sample_volume_uL = x
            elif 'min' in x and 'max' in x:
                target = f"between {x['min']}–{x['max']} µL"
                skip = v and x['min'] <= v <= x['max']
            elif 'min' in x:
                target = f"at least {x['min']} µL"
                skip = v and x['min'] <= v
            elif 'max' in x:
                target = f"at most {x['max']} µL"
                skip = v and v <= x['max']

            if not skip:
                ul += f"Ensure that the sample is {target}."

        ## Bind
        bind_params = zip_params(
                self.bind_buffer,
                self.bind_volume_x,
                self.bind_volume_uL,
        )
        for bind_buffer, bind_volume_x, bind_volume_uL in bind_params:
            bind_volume = resolve_volume(bind_volume_uL, bind_volume_x, self.sample_volume_uL)
            ul += f"Add {bind_volume} {bind_buffer} to the crude {self.sample_type}."

        if self.ph_buffer:
            ph_volume = resolve_volume(self.ph_volume_uL, self.ph_volume_x, self.sample_volume_uL)
            ul += f"If not {self.ph_color}: Add {ph_volume} {self.ph_buffer}."

        ul += f"Load on a {self.column_name}."
        ul += flush_column(self.bind_spin_sec, self.bind_vacuum)
        ul = break_if_too_long(pl, ul)

        ## Wash
        wash_params = zip_params(
                self.wash_buffer,
                self.wash_volume_uL,
                self.wash_spin_sec,
                self.wash_vacuum,
        )
        for wash_buffer, wash_volume_uL, wash_spin_sec, wash_vacuum in wash_params:
            ul += f"Add {wash_volume_uL} µL {wash_buffer}."
            ul += flush_column(wash_spin_sec, wash_vacuum)

        ## Dry
        if self.dry_spin_sec:
            ul += flush_column(self.dry_spin_sec)

        ul = break_if_too_long(pl, ul)

        ## Elute
        if self.elute_volume_uL < self.elute_min_volume_uL:
            warn(f"Elution volume ({self.elute_volume_uL} µL) is below the recommended minimum ({self.elute_min_volume_uL} µL).")

        ul += f"Add {self.elute_volume_uL} µL {self.elute_buffer}."
        if self.elute_wait_sec:
            ul += f"Wait at least {format_sec(self.elute_wait_sec)}."
        ul += flush_column(self.elute_spin_sec, keep_flowthrough=True)

        return p

def zip_params(*params):
    from itertools import repeat
    from more_itertools import always_iterable

    yield from zip(*(
            always_iterable(p or repeat(p))
            for p in params
    ))

def resolve_volume(volume_uL, volume_x, sample_volume_uL):
    if volume_uL:
        return f'{volume_uL} µL'
    elif sample_volume_uL:
        return f'{volume_x * sample_volume_uL} µL'
    else:
        return f'{volume_x} volumes'

def flush_column(spin_time_sec, use_vacuum=False, keep_flowthrough=False):
    if use_vacuum:
        return "Apply vacuum."
    else:
        if not spin_time_sec:
            raise ValueError("no spin time specified")
        return f"Spin {format_sec(spin_time_sec)}; {'keep' if keep_flowthrough else 'discard'} flow-through."


if __name__ == '__main__':
    SpinCleanup.main()




#!/usr/bin/env python3

import stepwise, appcli, autoprop
from inform import plural
from stepwise_mol_bio import Main, comma_set
from collections.abc import Sequence

def float_pair(x):
    if ',' in x:
        return tuple(map(float, x.split(',', maxsplit=1)))
    else:
        return float(x)

def pair(x, i):
    return x[i] if isinstance(x, Sequence) else x

@autoprop
class Anneal(Main):
    """\
Anneal two complementary oligos.

Usage:
    anneal <oligo_1> <oligo_2> [-n <num_rxns>] [-v <µL>] [options]

Arguments:
    <oligo_1> <oligo_2>
        The names of the two oligos to anneal.

Options:
    -n --num-rxns <num_rxns>            [default: ${app.num_reactions}]
        The number of reactions to set up.

    -v --volume <µL>                    [default: ${app.volume_uL}]
        The volume of each annealing reaction in µL.

    -c --oligo-conc <µM>
        The final concentration of each oligo in the reaction, in µM.  This 
        will also be the concentration of the annealed duplex, if the reaction 
        goes to completion.  The default is to use as much oligo as possible.

    -C --oligo-stock <µM[,µM]>          [default: ${app.oligo_stock_uM}]
        The stock concentrations of the oligos, in µM.  You can optionally use 
        a comma to specify different stock concentrations for the two oligos.

    -m --master-mix <reagents>          [default: ${','.join(app.master_mix)}]
        The reagents to include in the master mix.  The following reagents are 
        understood: '1' (the first oligo), '2' (the second oligo), or the name 
        of either oligo specified on the command line. To specify both 
        reagents, separate the two names with a comma.
"""
    __config__ = [
            appcli.DocoptConfig(),
    ]
    oligo_1 = appcli.param('<oligo_1>')
    oligo_2 = appcli.param('<oligo_2>')
    num_reactions = appcli.param('--num-rxns', default=1, cast=int)
    volume_uL = appcli.param('--volume', default=4, cast=float)
    oligo_conc_uM = appcli.param('--oligo-conc', default=None, cast=float)
    oligo_stock_uM = appcli.param('--oligo-stock', default=100, cast=float_pair)
    master_mix = appcli.param('--master-mix', default={'1'}, cast=comma_set)

    def __init__(self, oligo_1, oligo_2):
        self.oligo_1 = oligo_1
        self.oligo_2 = oligo_2

    def get_reaction(self):
        rxn = stepwise.MasterMix.from_text("""\
            Reagent  Stock     Volume  MM?
            =======  =====  =========  ===
            water           to 4.0 µL  yes
            PBS      10x       0.4 µL  yes
        """)
        rxn.num_reactions = self.num_reactions
        rxn.hold_ratios.volume = self.volume_uL, 'µL'

        rxn['oligo1'].name = self.oligo_1
        rxn['oligo2'].name = self.oligo_2
        rxn['oligo1'].master_mix = bool({'1', self.oligo_1} & self.master_mix)
        rxn['oligo2'].master_mix = bool({'2', self.oligo_2} & self.master_mix)
        rxn['oligo1'].stock_conc = pair(self.oligo_stock_uM, 0), 'µM'
        rxn['oligo2'].stock_conc = pair(self.oligo_stock_uM, 1), 'µM'

        if self.oligo_conc_uM:
            rxn['oligo1'].hold_stock_conc.conc = self.oligo_conc_uM, 'µM'
            rxn['oligo2'].hold_stock_conc.conc = self.oligo_conc_uM, 'µM'

        else:
            V = rxn.get_free_volume_excluding('oligo1', 'oligo2')
            C1 = rxn['oligo1'].stock_conc
            C2 = rxn['oligo2'].stock_conc
            C12 = C1 + C2

            rxn['oligo1'].volume = V * (C2 / C12)
            rxn['oligo2'].volume = V * (C1 / C12)

        return rxn

    def get_protocol(self):
        protocol = stepwise.Protocol()
        n = self.num_reactions

        protocol += f"""\
Setup {plural(n):# annealing reaction/s}:

{self.reaction}
"""
        protocol += f"""\
Perform the {plural(n):annealing reaction/s}:

- Incubate at 95°C for 2 min.
- Cool at room temperature.
"""
        return protocol

if __name__ == '__main__':
    Anneal.main()



#!/usr/bin/env python3

import pytest
import parametrize_from_file

from stepwise_mol_bio.pcr import Pcr, find_amplicon
from pytest import approx
from schema_helpers import *

@parametrize_from_file(
        schema=Schema({
            str: And(str, str.strip),
            'is_linear': eval,
            **error_or({
                'expected': And(str, str.strip),
            }),
        }),
)
def test_find_amplicon(template, primer_1, primer_2, is_linear, expected, error):
    with error:
        amplicon = find_amplicon(template, primer_1, primer_2, is_linear)
        assert amplicon == expected.upper()

@parametrize_from_file(schema=app_expected_error)
def test_product_seqs(app, expected, error):
    with error:
        assert app.product_seqs == [x.upper() for x in expected]

@parametrize_from_file(schema=app_expected)
def test_anneal_temp_C(app, expected):
    assert app.anneal_temp_C == approx(expected)

@parametrize_from_file(schema=app_expected)
def test_extend_time_s(app, expected):
    assert app.extend_time_s == approx(expected)

@parametrize_from_file(schema=app_expected)
def test_master_mix(app, expected):
    assert app.master_mix == expected

@parametrize_from_file(schema=db_expected)
def test_make(db, expected):
    for tag in expected:
        assert db[tag].seq == expected[tag]['seq'].upper()
        assert db[tag].dependencies == expected[tag]['dependencies']
        assert db[tag].conc_ng_uL == expected[tag]['conc_ng_uL']
        assert db[tag].molecule == 'DNA'
        assert db[tag].is_double_stranded == True
        assert db[tag].is_circular == False




#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import Key, DocoptConfig
from stepwise_mol_bio import Main, Pcr, Kld

@autoprop
class InversePcr(Pcr):
    """\
Clone a plasmid by inverse PCR.

Usage:
    {pcr}

    -L --skip-kld
        Skip the KLD reaction.  This is equivalent to just using the `pcr` 
        command.

    -P --skip-pcr
        Skip the PCR reaction.  This is equivalent to using the `kld` command 
        with a transformation step afterwards.
"""
    skip_pcr = appcli.param(
            Key(DocoptConfig, '--skip-pcr'),
            default=False,
    )
    skip_kld = appcli.param(
            Key(DocoptConfig, '--skip-kld'),
            default=False,
    )

    def __bareinit__(self):
        self.kld = Kld('PCR product')

    def format_usage(self):
        return self.__doc__.format(
                pcr=super().__doc__
                        .split('Usage:', 1)[1]
                        .replace('pcr', 'invpcr', 1)
                        .strip(),
        )

    def get_protocol(self):
        self.kld.num_reactions = self.num_reactions
        p = stepwise.Protocol()
        if not self.skip_pcr:
            p += super().protocol
        if not self.skip_kld:
            p += self.kld.protocol
            p += "Transform 2 µL."
        return p

    @property
    def is_product_circular(self):
        return True


if __name__ == '__main__':
    InversePcr.main()







#!/usr/bin/env python3

"""
Protocols relating to molecular biology, e.g. PCR.
"""

__version__ = '1.20.0'

from ._utils import *
from ._assembly import Assembly
from .aliquot import Aliquot
from .anneal import Anneal
from .digest import RestrictionDigest
from .direct_dilution import DirectDilution
from .ethanol_precipitation import EthanolPrecipitation
from .gels.gel import Gel
from .gels.laser_scanner import LaserScanner
from .gels.stain import Stain
from .gels.uv_transilluminator import UvTransilluminator
from .gibson import Gibson
from .golden_gate import GoldenGate
from .ivt import Ivt
from .ivtt import InVitroTranslation
from .kld import Kld
from .ligate import Ligate
from .pcr import Pcr
from .serial_dilution import SerialDilution
from .spin_cleanup import SpinCleanup

# Avoid circular imports
from .invpcr import InversePcr
from .qpcr import Qpcr
from .page_purify import PagePurify

from pathlib import Path

class Plugin:
    protocol_dir = Path(__file__).parent
    config_defaults = protocol_dir / 'conf.toml'

del Path



#!/usr/bin/env python3

import stepwise, appcli, autoprop
from stepwise import (
        StepwiseConfig, PresetConfig, MasterMix,
        paragraph_list, unordered_list,
)
from stepwise_mol_bio import (
        Main, UsageError,
        merge_dicts, format_min,
)
from freezerbox import (
        ReagentConfig, MakerArgsConfig,
        parse_volume_uL, parse_time_m, parse_temp_C, convert_conc_unit,
        unanimous, group_by_identity, normalize_seq, join_lists,
)
from appcli import DocoptConfig, Key, Method
from inform import plural, warn

def parse_reaction(reaction_input):
    if isinstance(reaction_input, dict):
        return {k: MasterMix(v) for k, v in reaction_input.items()}
    else:
        return MasterMix(reaction_input)

def pick_by_short(values, is_short):
    if not isinstance(values, dict):
        return values
    else:
        return values['short' if is_short else 'long']

def affected_by_short(values):
    if not isinstance(values, dict):
        return False
    else:
        return values['short'] != values['long']

class TemplateConfig(ReagentConfig):
    tag_getter = lambda app: app.templates

@autoprop.cache
class Ivt(Main):
    """\
Synthesize RNA using an in vitro transcription reaction.

Usage:
    ivt <templates>... [options]

Arguments:
    <templates>
        The names to the DNA templates to transcribe.  If these names can be 
        looked up in a FreezerBox database, some parameters (e.g. template 
        length) will automatically be determined.

<%! from stepwise_mol_bio import hanging_indent %>\
Options:
    -p --preset <name>          [default: ${app.preset}]
        What default reaction parameters to use.  The following parameters are 
        currently available:

        ${hanging_indent(app.preset_briefs, 8)}

    -v --volume <µL>
        The volume of the transcription reaction, in µL.

    -C --template-stock <ng/µL>
        The stock concentration of the template DNA (in ng/µL).  By default, 
        the concentration specified in the preset will be used.  Changing this 
        option will not change the quantity of template added to the reaction 
        (the volume will be scaled proportionally).

    -t --template-mass <ng>
        How much template DNA to use (in ng).  The template volume will be 
        scaled in order to reach the given quantity.  If the template is not 
        concentrated enough to reach the given quantity, the reaction will just 
        contain as much DNA as possible instead.  In order to use this option, 
        the stock concentration of the template must be in units of ng/µL.

    -V --template-volume <µL>
        The volume of DNA to add to the reaction (in µL).  This will be 
        adjusted if necessary to fit within the total volume of the reaction.

    -s --short
        Indicate that all of the templates are shorter than 300 bp.  This 
        allows the reactions to be setup with less material, so that more 
        reactions can be performed with a single kit.

    -i --incubation-time <minutes>
        How long to incubate the transcription reaction, in minutes.

    -T --incubation-temp <°C>
        What temperature the incubate the transcription reaction at, in °C.

    -x --extra <percent>        [default: ${app.extra_percent}]
        How much extra master mix to create.

    -R --toggle-rntp-mix
        Indicate whether you're using an rNTP mix, or whether you need to add 
        each rNTP individually to the reaction.  This option toggles the 
        default value specified in the configuration file.

    -a --toggle-dnase-treatment
        Indicate whether you'd like to include the optional DNase treatment 
        step.  This option toggles the default value specified in the 
        configuration file.

Configuration:
    Default values for this protocol can be specified in any of the following 
    stepwise configuration files:

        ${hanging_indent(app.config_paths, 8)}

    molbio.ivt.default_preset:
        The default value for the `--preset` option.

    molbio.ivt.rntp_mix:
        The default value for the `--toggle-rntp-mix` flag.

    molbio.ivt.dnase_treatment:
        The default value for the `--toggle-dnase-treament` flag.

    molbio.ivt.presets:
        Named groups of default reaction parameters.  Typically each preset 
        corresponds to a particular kit or protocol.  See below for the various 
        settings that can be specified in each preset.

    molbio.ivt.presets.<name>.brief:
        A brief description of the preset.  This is displayed in the usage info 
        for the `--preset` option.

    molbio.ivt.presets.<name>.inherit:
        Copy all settings from another preset.  This can be used to make small 
        tweaks to a protocol, e.g. "HiScribe with a non-standard additive".

    molbio.ivt.presets.<name>.reaction:
        A table detailing all the components of the transcription reaction, in 
        the format understood by `stepwise.MasterMix.from_string()`.  
        Optionally, this setting can be a dictionary with keys "long" and 
        "short", each corresponding to a reaction table.  This allows different 
        reaction parameters to be used for long and short templates.

        The DNA template reagent must be named "template".  The rNTP reagents 
        may be marked with the "rntp" flag.  The `--rntp-mix` flag will replace 
        any reagents so marked with a single reagent named "rNTP mix".

    molbio.ivt.presets.<name>.instructions:
        A list of miscellaneous instructions pertaining to how the reaction 
        should be set up, e.g. how to thaw the reagents, what temperature to 
        handle the reagents at, etc.

    molbio.ivt.presets.<name>.extra_percent:
        How much extra master mix to make, as a percentage of the volume of a 
        single reaction.

    molbio.ivt.presets.<name>.incubation_time_min:
        See `--incubation-time`.  This setting can also be a dictionary with 
        keys "long" and "short", to specify different incubation times for long 
        and short templates.

    molbio.ivt.presets.<name>.incubation_temp_C:
        See `--incubation-temp`.

    molbio.ivt.presets.<name>.length_threshold
        The length of a template in base pairs that separates long from short 
        templates. Template lengths will be queried from the FreezerBox 
        database if possible.

    molbio.ivt.presets.<name>.dnase.default
        The default value for the `--dnase` flag.

    molbio.ivt.presets.<name>.dnase.reaction
        A table detailing all the components of the optional DNase reaction.  
        One component must be named "transcription reaction".

    molbio.ivt.presets.<name>.dnase.incubation_time_min
        The incubation time (in minutes) for the optional DNase reaction.

    molbio.ivt.presets.<name>.dnase.incubation_temp_C
        The incubation temperature (in °C) for the optional DNase reaction.

    molbio.ivt.presets.<name>.dnase.footnotes.reaction
    molbio.ivt.presets.<name>.dnase.footnotes.incubation
    molbio.ivt.presets.<name>.dnase.footnotes.dnase
        Lists of footnotes for the reaction setup, incubation, and DNase 
        treatment steps, respectively.

Database:
    In vitro transcription reactions can appear in the "Synthesis" column of a 
    FreezerBox database.  The associated database entry will automatically be 
    considered ssRNA, e.g. for the purpose of molecular weight calculations:

        ivt template=<tag> [preset=<preset>] [volume=<µL>] [time=<min>]
            [temp=<°C>]

    template=<tag>
        See `<templates>`.  Only one template can be specified.

    preset=<preset>
        See `--preset`.

    volume=<µL>
        See `--volume`.

    time=<min>
        See `--incubation-time`.

    temp=<°C>
        See `--incubation-temp`.

Template Preparation:
    The following information is taken directly from the HiScribe and 
    MEGAscript manuals:

    Plasmid Templates:

        To produce RNA transcript of a defined length, plasmid DNA must be 
        completely linearized with a restriction enzyme downstream of the 
        insert to be transcribed.  Circular plasmid templates will generate 
        long heterogeneous RNA transcripts in higher quantities because of high 
        processivity of T7 RNA polymerase.  Be aware that there has been one 
        report of low level transcription from the inappropriate template 
        strand in plasmids cut with restriction enzymes leaving 3' overhanging 
        ends [Schendorn and Mierindorf, 1985].

        DNA from some miniprep procedures may be contaminated with residual 
        RNase A.  Also, restriction enzymes occasionally introduce RNase or 
        other inhibitors of transcription.  When transcription from a template 
        is suboptimal, it is often helpful to treat the template DNA with 
        proteinase K (100–200 μg/mL) and 0.5% SDS for 30 min at 50°C, follow 
        this with phenol/chloroform extraction (using an equal volume) and 
        ethanol precipitation.

    PCR Templates:

        PCR products containing T7 RNA Polymerase promoter in the correct 
        orientation can be transcribed.  Though PCR mixture can be used 
        directly, better yields will be obtained with purified PCR products.  
        PCR products can be purified according to the protocol for plasmid 
        restriction digests above, or by using commercially available spin 
        columns (we recommend Monarch PCR & DNA Cleanup Kit, NEB #T1030).  PCR 
        products should be examined on an agarose gel to estimate concentration 
        and to confirm amplicon size prior to its use as a template.  Depending 
        on the PCR products, 0.1–0.5 μg of PCR fragments can be used in a 20 μL 
        in vitro transcription reaction.

    Synthetic DNA Oligonucleotides:

        Synthetic DNA oligonucleotides which are either entirely 
        double-stranded or mostly single-stranded with a double-stranded T7 
        promoter sequence can be used for transcription.  In general, the 
        yields are relatively low and also variable depending upon the 
        sequence, purity, and preparation of the synthetic oligonucleotides.
"""
    __config__ = [
            DocoptConfig(),
            MakerArgsConfig(),
            TemplateConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.ivt'),
    ]
    preset_briefs = appcli.config_attr()
    config_paths = appcli.config_attr()

    def _calc_short(self):
        return all(
                x <= self.template_length_threshold
                for x in self.template_lengths
        )

    def _pick_by_short(self, values):
        return pick_by_short(values, self.short)

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(MakerArgsConfig, 'preset'),
            Key(StepwiseConfig, 'default_preset'),
    )
    reaction_prototype = appcli.param(
            Key(PresetConfig, 'reaction', cast=parse_reaction),
            get=_pick_by_short,
    )
    templates = appcli.param(
            Key(DocoptConfig, '<templates>'),
            Key(MakerArgsConfig, 'template', cast=lambda x: [x]),
    )
    template_seqs = appcli.param(
            Key(TemplateConfig, 'seq'),
    )
    template_lengths = appcli.param(
            Key(TemplateConfig, 'length'),
    )
    template_length_threshold = appcli.param(
            Key(PresetConfig, 'length_threshold'),
    )
    template_volume_uL = appcli.param(
            Key(DocoptConfig, '--template-volume', cast=float),
            default=None,
    )
    template_mass_ng = appcli.param(
            Key(DocoptConfig, '--template-mass', cast=float),
            default=None,
    )
    template_stock_ng_uL = appcli.param(
            Key(DocoptConfig, '--template-stock', cast=float),
            Key(TemplateConfig, 'conc_ng_uL', cast=min),
            default=None,
    )
    short = appcli.param(
            Key(DocoptConfig, '--short'),
            Method(_calc_short),
            default=False,
    )
    volume_uL = appcli.param(
            Key(DocoptConfig, '--volume-uL', cast=float),
            Key(MakerArgsConfig, 'volume', cast=parse_volume_uL),
            Key(PresetConfig, 'volume_uL'),
            default=None,
    )
    rntp_mix = appcli.toggle_param(
            Key(DocoptConfig, '--no-rntp-mix', toggle=True),
            Key(StepwiseConfig, 'rntp_mix'),
            default=True,
    )
    extra_percent = appcli.param(
            Key(DocoptConfig, '--extra-percent'),
            Key(PresetConfig, 'extra_percent'),
            cast=float,
            default=10,
    )
    instructions = appcli.param(
            Key(PresetConfig, 'instructions'),
            default_factory=list,
    )
    incubation_times_min = appcli.param(
            Key(DocoptConfig, '--incubation-time'),
            Key(MakerArgsConfig, 'time', cast=parse_time_m),
            Key(PresetConfig, 'incubation_time_min'),
    )
    incubation_temp_C = appcli.param(
            Key(DocoptConfig, '--incubation-temp'),
            Key(MakerArgsConfig, 'temp', cast=parse_temp_C),
            Key(PresetConfig, 'incubation_temp_C'),
    )
    dnase = appcli.toggle_param(
            Key(DocoptConfig, '--toggle-dnase-treatment', toggle=True),
            Key(PresetConfig, 'dnase.treatment'),
            Key(StepwiseConfig, 'dnase_treatment'),
            default=False,
    )
    dnase_reaction_prototype = appcli.param(
            Key(PresetConfig, 'dnase.reaction', cast=MasterMix),
    )
    dnase_incubation_time_min = appcli.param(
            Key(PresetConfig, 'dnase.incubation_time_min'),
    )
    dnase_incubation_temp_C = appcli.param(
            Key(PresetConfig, 'dnase.incubation_temp_C'),
    )
    footnotes = appcli.param(
            Key(PresetConfig, 'footnotes'),
            default_factory=dict,
    )

    group_by = {
        'preset': group_by_identity,
        'volume_uL': group_by_identity,
        'incubation_times_min': group_by_identity,
        'incubation_temp_C': group_by_identity,
    }
    merge_by = {
        'templates': join_lists,
    }

    def __init__(self, templates):
        self.templates = templates

    def __repr__(self):
        return f'Ivt(templates={self.templates!r})'

    def get_reaction(self):
        rxn = self.reaction_prototype.copy()
        rxn.num_reactions = len(self.templates)
        rxn.extra_percent = self.extra_percent

        if self.volume_uL:
            rxn.hold_ratios.volume = self.volume_uL, 'µL'

        if self.rntp_mix:
            rntps = []
            
            for i, reagent in enumerate(rxn):
                reagent.order = i
                if 'rntp' in reagent.flags:
                    rntps.append(reagent)

            if not rntps:
                err = ConfigError("cannot make rNTP mix", preset=self.preset)
                err.blame += "no reagents flagged as 'rntp'"
                err.hints += "you may need to add this information to the [molbio.ivt.{preset}] preset"
                raise err

            rxn['rNTP mix'].volume = sum(x.volume for x in rntps)
            rxn['rNTP mix'].stock_conc = sum(x.stock_conc for x in rntps) / len(rntps)
            rxn['rNTP mix'].master_mix = all(x.master_mix for x in rntps)
            rxn['rNTP mix'].order = rntps[0].order

            for rntp in rntps:
                del rxn[rntp.name]

        rxn['template'].name = ','.join(self.templates)

        if self.template_stock_ng_uL:
            rxn['template'].hold_conc.stock_conc = self.template_stock_ng_uL, 'ng/µL'

        if self.template_mass_ng:
            ng_uL = convert_conc_unit(rxn['template'].stock_conc, None, 'ng/µL')
            rxn['template'].volume = self.template_mass_ng / ng_uL.value, 'µL'

        if self.template_volume_uL:
            rxn['template'].volume = self.template_volume_uL, 'µL'
            if self.template_mass_ng:
                warn(f"template quantity ({self.template_mass_ng} ng) specified but overridden by volume ({self.template_volume_uL} µL)")
        
        rxn.fix_volumes('template', rxn.solvent)
        return rxn

    def get_dnase_reaction(self):
        rxn = self.dnase_reaction_prototype
        rxn.num_reactions = len(self.templates)
        rxn.extra_percent = self.extra_percent

        if self.volume_uL:
            k = (self.volume_uL, 'µL') / rxn['transcription reaction'].volume
            rxn.hold_ratios.volume *= k

        return rxn

    def get_protocol(self):
        p = stepwise.Protocol()

        ## Clean your bench
        p += stepwise.load('rnasezap')

        ## In vitro transcription
        rxn = self.reaction
        n = plural(rxn.num_reactions)
        f = self.footnotes.get('reaction', [])
        p += paragraph_list(
                f"Setup {n:# in vitro transcription reaction/s}{p.add_footnotes(*f)}:",
                rxn,
                unordered_list(*self.instructions),
        )

        f = self.footnotes.get('incubation', [])
        if self.short and affected_by_short(self.incubation_times_min):
            f += [f"Reaction time is different than usual because the template is short (<{self.template_length_threshold} bp)."]
        p += f"Incubate at {self.incubation_temp_C}°C for {format_min(pick_by_short(self.incubation_times_min, self.short))}{p.add_footnotes(*f)}."

        ## DNase treatment
        if self.dnase:
            f = self.footnotes.get('dnase', [])
            p += paragraph_list(
                    f"Setup {n:# DNase reaction/s}{p.add_footnotes(*f)}:",
                    self.dnase_reaction,
            )
            p += f"Incubate at {self.dnase_incubation_temp_C}°C for {format_min(self.dnase_incubation_time_min)}."

        return p
    def get_dependencies(self):
        return self.templates

    def get_product_seqs(self):
        return [transcribe(x) for x in self.template_seqs]

    def get_product_molecule(self):
        return 'ssRNA'

def transcribe(template_seq):
    t7_promoter = 'TAATACGACTCACTATA'
    template_seq = normalize_seq(template_seq)

    i = template_seq.find(t7_promoter)
    j = i + len(t7_promoter)

    if i < 0:
        raise ValueError(f"no T7 promoter found")

    return template_seq[j:].translate(str.maketrans('Tt', 'Uu'))


if __name__ == '__main__':
    Ivt.main()




#!/usr/bin/env python3

import stepwise, appcli, autoprop
from inform import Error
from stepwise_mol_bio.serial_dilution import SerialDilution, format_quantity

@autoprop
class DirectDilution(SerialDilution):
    """\
Dilute the given reagent in such a way that serial dilutions are minimized.

While serial dilutions are easy to perform, they have some disadvantages: 
First, they can be inaccurate, because errors made at any step are propagated 
to all other steps.  Second, they can waste material, because they produce more 
of the lowest dilution than is needed.  If either of these disadvantages are a 
concern, use this protocol to create the same dilution with fewer serial steps.

Usage:
    direct_dilution <volume> <high> to <low> <steps> [options]
    direct_dilution <volume> <high> / <factor> <steps> [options]
    direct_dilution <volume> <low> x <factor> <steps> [options]

Arguments:
    <volume>
        The volume of reagent needed for each concentration.  A unit may be 
        optionally given, in which case it will be included in the protocol.

    <high>
        The starting concentration for the dilution.  A unit may be optionally 
        given, in which case it will be included in the protocol.

    <low>
        The ending concentration for the dilution.  A unit may be optionally 
        given, in which case it will be included in the protocol.

    <factor>
        How big of a dilution to make at each step of the protocol.

    <steps>
        The number of dilutions to make, including <high> and <low>.

Options:
    -m --material <name>        [default: ${app.material}]
        The substance being diluted.

    -d --diluent <name>         [default: ${app.diluent}]
        The substance to dilute into.

    -x --max-dilution <fold>    [default: ${app.max_dilution}]
        Specify the biggest dilution that can be made at any step, as larger 
        dilutions are prone to be less accurate.

    -0 --include-zero
        Include a "dilution" with no material in the protocol.
"""
    max_dilution = appcli.param('--max-dilution', cast=float, default=10)

    def get_protocol(self):
        header = [
                format_quantity('Final', f'[{self.conc_unit}]', pad='\n'),
                format_quantity('Stock', f'[{self.conc_unit}]', pad='\n'),
                format_quantity(self.material, f'[{self.volume_unit}]', pad='\n'),
                format_quantity(self.diluent, f'[{self.volume_unit}]', pad='\n'),
        ]
        rows = []
        volumes = {}
        stock_concs = self._pick_stock_concs()

        for target_conc in reversed(self.concentrations):
            target_volume = self.volume + volumes.get(target_conc, 0)
            stock_conc = stock_concs[target_conc]
            stock_volume = target_volume * target_conc / stock_conc

            volumes[stock_conc] = volumes.get(stock_conc, 0) + stock_volume
            rows.insert(0, [
                format_quantity(target_conc),
                format_quantity(stock_conc),
                format_quantity(stock_volume),
                format_quantity(target_volume - stock_volume),
            ])

        protocol = stepwise.Protocol()
        protocol += f"""\
Prepare the following dilutions:

{stepwise.tabulate(rows, header, align='>>>>')}
"""
        return protocol

    def _pick_stock_concs(self):
        stock_concs = {}
        stock_conc = self.conc_high
        prev_target_conc = None

        for target_conc in self.concentrations:
            if target_conc == 0:
                stock_concs[target_conc] = stock_conc
                break

            dilution = stock_conc / target_conc
            if dilution > self.max_dilution:
                stock_conc = prev_target_conc

            dilution = stock_conc / target_conc
            if dilution > self.max_dilution:
                raise Error(f"{dilution:.1g}x dilution to make {format_quantity(target_conc, self.conc_unit)} exceeds maximum ({self.max_dilution:.1g}x)")

            stock_concs[target_conc] = stock_conc
            prev_target_conc = target_conc

        return stock_concs

if __name__ == '__main__':
    DirectDilution.main()

# vim: tw=53



#!/usr/bin/env python3

import pytest
import parametrize_from_file
from stepwise_mol_bio.digest import *
from freezerbox import Database, Plasmid
from schema_helpers import *
from warnings import catch_warnings, simplefilter

with catch_warnings():
    simplefilter('ignore', DeprecationWarning)
    import requests_testing

@parametrize_from_file(
        schema=Schema({
            'seq': eval,
            'enzymes': eval,
            'is_circular': eval,
            'target_size': eval,
            **error_or({
                'product': eval,
                'products': eval,
            }),
        }),
)
def test_calc_digest_products(seq, enzymes, is_circular, target_size, product, products, error):
    with error:
        assert products == calc_digest_products(
                seq, enzymes,
                is_circular=is_circular,
        )
    with error:
        assert product == calc_digest_product(
                seq, enzymes,
                is_circular=is_circular,
                target_size=target_size,
        )

def test_neb_restriction_enzyme_database(tmp_path):
    cache_path = tmp_path / 'cache.json'
    db = NebRestrictionEnzymeDatabase(cache_path)

    # This doesn't test the case where the internet is inaccessible.

    assert db['EcoRI'] == db['ecori'] == {
            'amt': '50000/10000/50000/10000 units',
            'bcn': 'R0101',
            'blueWhite': True,
            'buf1': 25,
            'buf2': 100,
            'buf3': 50,
            'buf4': 50,
            'buf5': 0,
            'cat': 'R0101',
            'clonedAtNEB': True,
            'concentration': 20000,
            'cpg': True,
            'dam': False,
            'dcm': False,
            'displayName': 'EcoRI',
            'engineered': False,
            'epimarkValidated': False,
            'heatInactivationTemp': 65,
            'heatInactivationTime': 20,
            'hfEnzyme': False,
            'incubateTemp': 37,
            'methylationSensitivity': [
                'cpg (Blocked by Some Combinations of Overlapping)',
            ],
            'mul': False,
            'name': 'EcoRI',
            'plainname': 'EcoRI',
            'recombinant': True,
            'recommBuffer': 'NEBuffer EcoRI/SspI',
            'reducedStarActivity': False,
            'size': 'L/S/M/T ',
            'star1': False,
            'star2': True,
            'star3': False,
            'star4': True,
            'star5': False,
            'supplement': {'atp': 0.0, 'bsa': 0.0, 'dtt': 0.0, 'enzact': 0.0, 'sam': 0.0},
            'thermophilic': False,
            'timeSaver': True,
            'url': 'https://www.neb.com/products/r0101-ecori',
    }

    with pytest.raises(ConfigError) as err:
        db['EcoRJ']

    assert err.match(r"no such enzyme 'EcoRJ'")
    assert err.match(r"successfully downloaded the most recent restriction enzyme data from NEB \(in case 'EcoRJ' is a new enzyme\)")
    assert err.match(r"did you mean: 'EcoRI'")

@requests_testing.activate
def test_neb_restriction_enzyme_database_offline(tmp_path):
    cache_path = tmp_path / 'cache.json'

    with pytest.raises(ConfigError) as err:
        NebRestrictionEnzymeDatabase(cache_path)

    assert err.match("failed to download")
    assert err.match("URL: http://nebcloner.neb.com/data/reprop.json")

@parametrize_from_file(
        schema=Schema({
            'enzymes': eval_with(),
            'expected': str,
        }),
)
def test_pick_compatible_buffer(enzymes, expected):
    assert pick_compatible_buffer(enzymes) == expected

@parametrize_from_file(schema=app_expected_reaction)
def test_reaction(app, expected):
    assert app.reaction.format_text() == expected

def test_reaction_unknown_supplement():
    mock_db = {
            'XyzX': {
                'name': 'XyzX',
                'displayName': 'XyzX',
                'concentration': 10_000,
                'recommBuffer': 'rCutSmart Buffer',
                "supplement": {
                    "atp": 0.0,
                    "bsa": 0.0,
                    "dtt": 0.0,
                    "enzact": 20.0, 
                    "sam": 0.0,
                },
            },
    }
    app = RestrictionDigest(['x1'], ['XyzX'], mock_db)

    with pytest.raises(ConfigError) as err:
        app.reaction

    assert err.match(r"'XyzX' requires an unknown supplement: 'enzact'")

@parametrize_from_file(schema=app_expected_protocol)
def test_protocol(app, expected):
    actual = app.protocol.format_text()
    print(actual)
    for x in expected:
        assert x in actual

@parametrize_from_file(schema=db_expected)
def test_make(db, expected):
    for tag in expected:
        assert db[tag].seq == expected[tag]['seq']
        assert db[tag].dependencies == expected[tag]['dependencies']

